/*
copyright (c) 2008 Kazuki IWAMOTO http://www.maid.org/ iwm@maid.org

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


char *
x_strcat (char *str, ...)
{
  char *p;
  size_t len;
  va_list args;

  len = (str ? strlen (str) : 0) + 1;
  va_start (args, str);
  while (p = va_arg (args, char *))
    len += strlen (p);
  va_end (args);
  str = str ? realloc (str, len * sizeof (char)) : calloc (len, sizeof (char));
  va_start (args, str);
  while (p = va_arg (args, char *))
    strcat (str, p);
  va_end (args);
  return str;
}


char *
x_strrep_b (char *str,
            const char *pattern, ...)
{
  char *p;

  if (str && pattern && (p = strstr (str, pattern)))
    {
      char *q;

      q = p + strlen (pattern);
      if ((p == str || isspace (*(p - 1))) && (*q == '\0' || isspace (*q)))
        {
          char *r, *tmp;
          size_t len;
          va_list args;

          len = p - str + strlen (q) + 1;
          va_start (args, pattern);
          while (r = va_arg (args, char *))
            len += strlen (r);
          va_end (args);
          tmp = malloc (len * sizeof (char));
          *p = '\0';
          strcpy (tmp, str);
          va_start (args, pattern);
          while (r = va_arg (args, char *))
            strcat (tmp, r);
          va_end (args);
          strcat (tmp, q);
          free (str);
          str = tmp;
        }
    }
  return str;
}


char *
x_strrep_g (char *str,
            const char *pattern, ...)
{
  char *p;

  if (str && pattern)
    for (p = strstr (str, pattern); p; p = strstr (p, pattern))
      {
        char *q, *tmp;
        size_t len;
        va_list args;

        len = strlen (str) - strlen (pattern) + 1;
        va_start (args, pattern);
        while (q = va_arg (args, char *))
          len += strlen (q);
        va_end (args);
        tmp = malloc (len * sizeof (char));
        *p = '\0';
        strcpy (tmp, str);
        va_start (args, pattern);
        while (q = va_arg (args, char *))
          strcat (tmp, q);
        va_end (args);
        q = p + strlen (pattern);
        p = tmp + strlen (tmp);
        strcat (tmp, q);
        free (str);
        str = tmp;
      }
  return str;
}


char *
x_fgets (FILE *fp)
{
  char buf[BUFSIZ], *str = NULL;

  while (fgets (buf, BUFSIZ, fp))
    {
      char c = '\0';
      size_t len;

      len = strlen (buf);
      if (len > 0)
        {
          c = buf[len - 1];
          if (c == '\n')
            buf[--len] = '\0';
        }
      str = x_strcat (str, buf, NULL);
      if (c == '\n')
        break;
    }
  return str;
}


static size_t size;


char *
sub_mask (char *str)
{
  char *mask;

  mask = malloc ((2 + strlen (str) + (size > 8 ? 4 : 0) + 1) * sizeof (char));
  strcpy (mask, "0x");
  strcat (mask, str);
  if (size > 8)
    strcat (mask, "Ui64");
  free (str);
  return mask;
}


int
main (int   argc,
      char *argv[])
{
  char *mask, *zero_mask, *one_half, *str;
  char *g_mask, *b_mask, *a_mask, *rb_mask, *ag_mask;
  char *rb_one_half, *rb_mask_plus_one;
  char size1[BUFSIZ], size2[BUFSIZ], size3[BUFSIZ];
  char pixel_size1[BUFSIZ], half_pixel_size1[BUFSIZ];
  size_t pixel_size, half_pixel_size, nibbles;

  if (argc != 2)
    return -1;
  size = atoi (argv[1]);
  if (size != 8 && size != 16)
    return -1;

  pixel_size = size * 4;
  half_pixel_size = size * 2;

  nibbles = size / 4;
  mask = malloc ((nibbles + 1) * sizeof (char));
  memset (mask, 'f', nibbles);
  mask[nibbles] = '\0';
  zero_mask = malloc ((nibbles + 1) * sizeof (char));
  memset (zero_mask, '0', nibbles);
  zero_mask[nibbles] = '\0';
  one_half = malloc ((nibbles + 1) * sizeof (char));
  memset (one_half + 1, '0', nibbles - 1);
  one_half[0] = '8';
  one_half[nibbles] = '\0';

  fputs (
"/* WARNING: This file is generated by iwm-combine-win32.exe.\n"
"   Please edit one of those files rather than this one. */\n", stdout);

  g_mask = sub_mask (x_strcat (NULL, mask, zero_mask, NULL));
  b_mask = sub_mask (x_strcat (NULL, mask, zero_mask, zero_mask, NULL));
  a_mask = sub_mask (x_strcat (NULL,
                                mask, zero_mask, zero_mask, zero_mask, NULL));
  rb_mask = sub_mask (x_strcat (NULL, mask, zero_mask, mask, NULL));
  ag_mask = sub_mask (x_strcat (NULL, mask, zero_mask, mask, zero_mask, NULL));
  rb_one_half = sub_mask (x_strcat (NULL,
                                        one_half, zero_mask, one_half, NULL));
  rb_mask_plus_one = sub_mask (x_strcat (NULL,
                            "1", zero_mask, zero_mask, "1", zero_mask, NULL));
  mask = sub_mask (mask);
  one_half = sub_mask (one_half);

  sprintf (size1, "%d", size);
  sprintf (size2, "%d", size * 2);
  sprintf (size3, "%d", size * 3);
  sprintf (pixel_size1, "%d", pixel_size);
  sprintf (half_pixel_size1, "%d", half_pixel_size);

  while (str = x_fgets (stdin))
    {
      str = x_strrep_b (str, "#define COMPONENT_SIZE",
                             "#define COMPONENT_SIZE ", size1, NULL);
      str = x_strrep_b (str, "#define MASK",
                             "#define MASK ", mask, NULL);
      str = x_strrep_b (str, "#define ONE_HALF",
                             "#define ONE_HALF ", one_half, NULL);

      str = x_strrep_b (str, "#define G_SHIFT",
                             "#define G_SHIFT ", size1, NULL);
      str = x_strrep_b (str, "#define B_SHIFT",
                             "#define B_SHIFT ", size2, NULL);
      str = x_strrep_b (str, "#define A_SHIFT",
                             "#define A_SHIFT ", size3, NULL);
      str = x_strrep_b (str, "#define G_MASK",
                             "#define G_MASK ", g_mask, NULL);
      str = x_strrep_b (str, "#define B_MASK",
                             "#define B_MASK ", b_mask, NULL);
      str = x_strrep_b (str, "#define A_MASK",
                             "#define A_MASK ", a_mask, NULL);

      str = x_strrep_b (str, "#define RB_MASK",
                             "#define RB_MASK ", rb_mask, NULL);
      str = x_strrep_b (str, "#define AG_MASK",
                             "#define AG_MASK ", ag_mask, NULL);
      str = x_strrep_b (str, "#define RB_ONE_HALF",
                             "#define RB_ONE_HALF ", rb_one_half, NULL);
      str = x_strrep_b (str, "#define RB_MASK_PLUS_ONE",
                        "#define RB_MASK_PLUS_ONE ", rb_mask_plus_one, NULL);

      str = x_strrep_b (str, "CombineFuncC",
                             "CombineFuncC", pixel_size1, NULL);
      str = x_strrep_b (str, "CombineFuncU",
                             "CombineFuncU", pixel_size1, NULL);
      str = x_strrep_b (str, "CombineMaskU",
                             "CombineMaskU", pixel_size1, NULL);
      str = x_strrep_b (str, "FbComposeFunctions",
                             "FbComposeFunctions", pixel_size1, NULL);

      str = x_strrep_g (str, "comp1_t",
                             "uint", size1, "_t", NULL);
      str = x_strrep_g (str, "comp2_t",
                             "uint", half_pixel_size1, "_t", NULL);
      str = x_strrep_g (str, "comp4_t",
                             "uint", pixel_size1, "_t", NULL);

      if (size == 16)
        str = x_strrep_g (str, "pixman_composeFunctions",
                               "pixman_composeFunctions64", NULL);

      if (size == 16)
        str = x_strrep_g (str, "pixman-combine.h",
                               "pixman-combine64.h", NULL);
      if (size == 8)
        str = x_strrep_g (str, "pixman-combine.h",
                               "pixman-combine32.h", NULL);

      fprintf (stdout, "%s\n", str);
      free (str);
    }

  return 0;
}
